# salutator/__init__.py
"""
Salutator — a package for greetings and goodbyes
across humans, animals, plants, and minerals.
"""

from . import humans
from . import animals
from . import plants
from . import minerals
from . import eso


#  Version info (__version__)
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0"


__all__ = ['humans', 'animals', 'plants', 'minerals']




# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
===========================
ESO Archive Module tests
===========================

European Southern Observatory (ESO)

"""

from collections import Counter
import pytest

from astropy.table import Table
from astroquery.exceptions import NoResultsWarning, MaxResultsWarning
from salutator.eso import Eso

instrument_list = ['alpaca', 'fors1', 'fors2', 'sphere', 'vimos', 'omegacam',
                   'hawki', 'isaac', 'naco', 'visir', 'vircam',
                   'apex',
                   'giraffe', 'uves', 'xshooter', 'muse', 'crires',
                   'kmos', 'sinfoni', 'amber', 'midi', 'pionier',
                   'gravity', 'espresso', 'wlgsu', 'matisse', 'eris',
                   'fiat',
                   'efosc', 'harps', 'nirps', 'sofi'
                   ]

SGRA_SURVEYS = ['195.B-0283',
                'ALMA',
                'ATLASGAL',
                'CRIRESplus',
                'ERIS-SPIFFIER',
                'GIRAFFE',
                'HARPS',
                'HAWKI',
                'KMOS',
                'MW-BULGE-PSFPHOT',
                'SEDIGISM',
                'VPHASplus',
                'VVV',
                'VVVX',
                'XSHOOTER'
                ]

ONE_RECORD_SURVEYS = [
    '081.C-0827',
    '108.2289',
    '1100.A-0528',
    '60.A-9493',
    'APEX-SciOps',
    'HARPS',
    'LESS',
]


@pytest.mark.remote_data
class TestEso:
    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_query_tap_service(self):
        eso = Eso()
        eso.ROW_LIMIT = 7
        with pytest.warns(MaxResultsWarning):
            t = eso.query_tap("select * from ivoa.ObsCore")
        lt = len(t)
        assert isinstance(t, Table), f"Expected type {type(Table)}; Obtained {type(t)}"
        assert len(t) > 0, "Table length is zero"
        assert len(t) == eso.ROW_LIMIT, f"Table length is {lt}, expected {eso.ROW_LIMIT}"

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_row_limit(self):
        eso = Eso()
        eso.ROW_LIMIT = 5
        # since in this case the results are truncated, a warning is issued

        with pytest.warns(MaxResultsWarning):
            table = eso.query_instrument('UVES')
            n = len(table)
            assert n == eso.ROW_LIMIT, f"Expected {eso.ROW_LIMIT}; Obtained {n}"

        with pytest.warns(MaxResultsWarning):
            table = eso.query_surveys('VVV')
            n = len(table)
            assert n == eso.ROW_LIMIT, f"Expected {eso.ROW_LIMIT}; Obtained {n}"

        with pytest.warns(MaxResultsWarning):
            table = eso.query_main()
            n = len(table)
            assert n == eso.ROW_LIMIT, f"Expected {eso.ROW_LIMIT}; Obtained {n}"

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_top(self):
        eso = Eso()
        top = 5
        eso.ROW_LIMIT = None
        # in this case the results are NOT truncated, no warnings should be issued

        table = eso.query_instrument('UVES', top=top)
        n = len(table)
        assert n == top, f"Expected {top}; Obtained {n}"

        table = eso.query_surveys('VVV', top=top)
        n = len(table)
        assert n == top, f"Expected {top}; Obtained {n}"

        table = eso.query_main(top=top)
        n = len(table)
        assert n == top, f"Expected {top}; Obtained {n}"

    def test_apex_retrieval(self):
        eso = Eso()

        tblb = eso.query_apex_quicklooks(
            column_filters={
                "project_id": 'E-095.F-9802A-2015'
            }
        )
        tbla = eso.query_apex_quicklooks(
            column_filters={
                "prog_id": '095.F-9802(A)'
            }
        )

        assert len(tbla) == 5
        assert set(tbla['release_date']) == {
            '2015-07-17T03:06:23.280Z',
            '2015-07-18T12:07:32.713Z',
            '2015-09-18T11:31:15.867Z',
            '2015-09-15T11:06:55.663Z',
            '2015-09-18T11:46:19.970Z'
        }

        assert all(tbla.values_equal(tblb))

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_sgrastar(self):
        eso = Eso()
        eso.ROW_LIMIT = 1

        instruments = eso.list_instruments()
        # in principle, we should run both of these tests
        # result_i = eso.query_instrument('midi', target='Sgr A*')
        # Equivalent, does not depend on SESAME:
        with pytest.warns(MaxResultsWarning):
            result_i = eso.query_instrument('midi', cone_ra=266.41681662,
                                            cone_dec=-29.00782497, cone_radius=1.0)

        surveys = eso.list_surveys()
        assert len(surveys) > 0

        with pytest.warns(MaxResultsWarning):
            result_s = eso.query_surveys(surveys='VVV', cone_ra=266.41681662,
                                         cone_dec=-29.00782497,
                                         cone_radius=1.0)

        assert 'midi' in instruments
        assert result_i is not None
        assert 'VVV' in surveys
        assert result_s is not None

        # From obs.raw, we have "object" (when query_instruments)
        # object: Target designation as given by the astronomer,
        # though at times overwritten by the obeservatory,
        # especially for CALIB observations. Compare with the similar field called "target".)

        # From ivoa.ObsCore, we have "target_name" (when query_surveys)
        # target_name: The target name as assigned by the Principal Investigator;
        # ref. Ref. OBJECT keyword in ESO SDP standard.
        # For spectroscopic public surveys, the value shall be set to the survey source identifier.
        assert 'target_name' in result_s.colnames
        assert 'b319' in result_s['target_name']

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_multisurvey(self, tmp_path):

        eso = Eso()
        eso.cache_location = tmp_path
        eso.ROW_LIMIT = 1000

        test_surveys = ['VVV', 'XSHOOTER']
        with pytest.warns(MaxResultsWarning):
            result_s = eso.query_surveys(surveys=test_surveys,
                                         cone_ra=266.41681662,
                                         cone_dec=-29.00782497,
                                         cone_radius=1.0)

        assert result_s is not None
        assert 'target_name' in result_s.colnames

        counts = Counter(result_s["obs_collection"].data)
        for tc in test_surveys:
            assert counts[tc] > 0, f"{tc} : survey not present in results"

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_empty_return(self):
        # test for empty return with an object from the North
        eso = Eso()
        surveys = eso.list_surveys()
        assert len(surveys) > 0

        # Avoid SESAME
        with pytest.warns(NoResultsWarning):
            result_s = eso.query_surveys(surveys=surveys[0], cone_ra=202.469575,
                                         cone_dec=47.195258, cone_radius=1.0)

        assert len(result_s) == 0

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_sgrastar_column_filters(self):
        eso = Eso()

        result1 = eso.query_surveys("sphere, vegas",
                                    columns=("obs_collection, calib_level, "
                                             "multi_ob, filter, s_pixel_scale, instrument_name"),
                                    column_filters={
                                        'calib_level': "= 3",
                                        'multi_ob': "like '%M%'"}
                                    )

        result2 = eso.query_surveys("sphere, vegas",
                                    columns=("obs_collection, calib_level, "
                                             "multi_ob, filter, s_pixel_scale, instrument_name"),
                                    column_filters={
                                        'calib_level': 3,
                                        'multi_ob': 'M'
                                    }
                                    )

        assert all(result1.values_equal(result2))

    def test_list_instruments(self):
        # If this test fails, we may simply need to update it
        inst = set(Eso.list_instruments())
        assert set(inst) == set(instrument_list), \
            f"Expected result {instrument_list}; Obtained: {inst}"

    def test_retrieve_data(self):
        eso = Eso()
        file_id = 'AMBER.2006-03-14T07:40:19.830'
        result = eso.retrieve_data(file_id)
        assert isinstance(result, str)
        assert file_id in result

    @pytest.mark.skipif('not Eso.USERNAME')
    def test_retrieve_data_authenticated(self):
        eso = Eso()
        eso.login()
        file_id = 'AMBER.2006-03-14T07:40:19.830'
        result = eso.retrieve_data(file_id)
        assert isinstance(result, str)
        assert file_id in result

    def test_retrieve_data_list(self):
        eso = Eso()
        datasets = ['MIDI.2014-07-25T02:03:11.561', 'AMBER.2006-03-14T07:40:19.830']
        result = eso.retrieve_data(datasets)
        assert isinstance(result, list)
        assert len(result) == 2

    @pytest.mark.parametrize('instrument', instrument_list)
    def test_help(self, instrument):
        eso = Eso()
        eso.query_instrument(instrument, help=True)

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    @pytest.mark.parametrize('instrument', instrument_list)
    def test_each_instrument_sgrastar(self, instrument):
        eso = Eso()
        eso.ROW_LIMIT = 1  # Either we have maxrec results or none at all
        try:
            with pytest.warns(MaxResultsWarning):
                result = eso.query_instrument(instrument,
                                              cone_ra=266.41681662,
                                              cone_dec=-29.00782497,
                                              cone_radius=1.0)
        except NoResultsWarning:  # we don't care if there are no results
            pass
        else:
            assert isinstance(result, Table)
            assert len(result) > 0, f"query_instrument({instrument}) returned no records"

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_each_survey_sgrastar(self, tmp_path):
        eso = Eso()
        eso.cache_location = tmp_path
        eso.ROW_LIMIT = 1

        surveys = eso.list_surveys()
        for survey in surveys:
            if survey in SGRA_SURVEYS:  # survey contains SGRA
                if survey not in ONE_RECORD_SURVEYS:  # Expect warnings
                    with pytest.warns(MaxResultsWarning):
                        result_s = eso.query_surveys(
                            surveys=survey,
                            cone_ra=266.41681662, cone_dec=-29.00782497, cone_radius=0.1775)
                else:  # No warnings expected
                    result_s = eso.query_surveys(
                        surveys=survey,
                        cone_ra=266.41681662, cone_dec=-29.00782497, cone_radius=0.1775)

                assert isinstance(result_s, Table)
                assert len(result_s) > 0
            else:  # survey does not contain SGRA
                with pytest.warns(NoResultsWarning):
                    result_s = eso.query_surveys(surveys=survey, cone_ra=266.41681662,
                                                 cone_dec=-29.00782497,
                                                 cone_radius=0.1775)
                    assert isinstance(result_s, Table)
                    assert isinstance(result_s, Table)
                    assert len(result_s) == 0, f"Failed for survey {survey}"

                if survey not in ONE_RECORD_SURVEYS:  # Expect warnings
                    with pytest.warns(MaxResultsWarning):
                        generic_result = eso.query_surveys(surveys=survey)

                else:  # Do not expect warnings
                    generic_result = eso.query_surveys(surveys=survey)
                assert isinstance(generic_result, Table)
                assert len(generic_result) > 0, \
                    f"query_surveys({survey}) returned no records"

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_mixed_case_instrument(self, tmp_path):
        eso = Eso()
        eso.cache_location = tmp_path
        eso.ROW_LIMIT = 5

        with pytest.warns(MaxResultsWarning):
            result1 = eso.query_instrument('midi', cone_ra=266.41681662,
                                           cone_dec=-29.00782497, cone_radius=1.0)
            result2 = eso.query_instrument('MiDi', cone_ra=266.41681662,
                                           cone_dec=-29.00782497, cone_radius=1.0)
        assert all(result1.values_equal(result2))

    @pytest.mark.filterwarnings("ignore::pyvo.dal.exceptions.DALOverflowWarning")
    def test_main_sgrastar(self):
        eso = Eso()
        eso.ROW_LIMIT = 5

        with pytest.warns(MaxResultsWarning):
            result = eso.query_main(
                column_filters={
                    'target': "SGR A",
                    'object': "SGR A"}
            )

        assert len(result) == 5
        assert 'SGR A' in result['object']
        assert 'SGR A' in result['target']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
===========================
ESO Archive Module tests
===========================

European Southern Observatory (ESO)

"""
import os
import shutil
import sys

import pytest
import pyvo
from astropy.table import Table
import astropy.io.ascii

from astroquery.utils.mocks import MockResponse
from ...eso import Eso
from ...eso.utils import _UserParams, \
    _build_adql_string, _adql_sanitize_op_val, _reorder_columns, \
    DEFAULT_LEAD_COLS_RAW
from astroquery.exceptions import NoResultsWarning, MaxResultsWarning

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
EXPECTED_MAXREC = 1000
MONKEYPATCH_TABLE_LENGTH = 50


def data_path(filename):
    return os.path.join(DATA_DIR, filename)


DATA_FILES = {
    'GET':
        {
            Eso.AUTH_URL: 'oidc_token.json',
        },
    'POST':
        {
            'https://archive.eso.org/wdb/wdb/eso/eso_archive_main/query': 'main_sgra_query.tbl',
            'https://archive.eso.org/wdb/wdb/eso/amber/query': 'amber_sgra_query.tbl',
            'https://archive.eso.org/wdb/wdb/adp/phase3_main/query': 'vvv_sgra_survey_response.tbl',
        },
    'ADQL':
        {
            "select * from ivoa.ObsCore where obs_collection in ('VVV') and "
            "intersects(s_region, circle('ICRS', 266.41681662, -29.00782497, 0.1775))=1":
            "query_coll_vvv_sgra.csv",

            "select * from ist.sinfoni where target = 'SGRA'":
            "query_inst_sinfoni_sgra.csv",

            "select * from ist.apex_quicklooks where project_id = 'E-095.F-9802A-2015'":
            "query_apex_ql_5.csv",

            "select * from dbo.raw where target = 'SGR A' and object = 'SGR A'":
            "query_main_sgra.csv",

            "select distinct obs_collection from ivoa.ObsCore": "query_list_surveys.csv",

            "select table_name from TAP_SCHEMA.tables where schema_name='ist' order by table_name":
            "query_list_instruments.csv",
        }
}

TEST_SURVEYS = [
    '081.C-0827', 'ADHOC', 'CAFFEINE', 'ENTROPY', 'GAIAESO', 'HARPS', 'INSPIRE', 'KIDS', 'ZCOSMOS']
TEST_INSTRUMENTS = [
    'amber', 'crires', 'espresso', 'fors1', 'giraffe', 'gravity', 'midi', 'xshooter']


def eso_request(request_type, url, **kwargs):
    _ = kwargs
    with open(data_path(DATA_FILES[request_type][url]), 'rb') as f:
        response = MockResponse(content=f.read(), url=url)
    return response


def monkey_tap(query, **kwargs):
    _ = kwargs
    table_file = data_path(DATA_FILES['ADQL'][query])
    table = astropy.io.ascii.read(table_file, format='csv', header_start=0, data_start=1)
    return table


def download_request(url, **kwargs):
    _ = kwargs
    filename = 'testfile.fits.Z'
    with open(data_path(filename), 'rb') as f:
        header = {'Content-Disposition': f'filename={filename}'}
        response = MockResponse(content=f.read(), url=url, headers=header)
    return response


def calselector_request(url, **kwargs):
    is_multipart = len(kwargs['data']['dp_id']) > 1
    if is_multipart:
        filename = 'FORS2.2021-01-02T00_59_12.533_raw2raw_multipart.xml'
        header = {
            'Content-Type': 'multipart/form-data; '
            'boundary=uFQlfs9nBIDEAIoz0_ZM-O2SXKsZ2iSd4h7H;charset=UTF-8'
        }
    else:
        filename = 'FORS2.2021-01-02T00_59_12.533_raw2raw.xml'
        header = {
            'Content-Disposition': f'filename="{filename}"',
            'Content-Type': 'application/xml; content=calselector'
        }
    with open(data_path(filename), 'rb') as f:
        response = MockResponse(content=f.read(), url=url, headers=header)
    return response


def test_sinfoni_sgr_a_star(monkeypatch):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    result = eso.query_instrument('sinfoni',
                                  column_filters={
                                      'target': "SGRA"
                                  }
                                  )
    # test all results are there and the expected target is present
    assert len(result) == MONKEYPATCH_TABLE_LENGTH
    assert 'SGRA' in result['target']


def test_main_sgr_a_star(monkeypatch):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    result = eso.query_main(
        column_filters={
            'target': "SGR A",
            'object': "SGR A"
        })
    # test all results are there and the expected target is present
    assert len(result) == 23
    assert 'SGR A' in result['object']
    assert 'SGR A' in result['target']


def test_apex_retrieval(monkeypatch):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)

    tbla = eso.query_apex_quicklooks(
        column_filters={
            "project_id": 'E-095.F-9802A-2015'
        }
    )

    assert len(tbla) == 5
    assert set(tbla['release_date']) == {
        '2015-07-17T03:06:23.280Z',
        '2015-07-18T12:07:32.713Z',
        '2015-09-18T11:31:15.867Z',
        '2015-09-15T11:06:55.663Z',
        '2015-09-18T11:46:19.970Z'
    }


def test_vvv(monkeypatch):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    result = eso.query_surveys(surveys='VVV',
                               cone_ra=266.41681662, cone_dec=-29.00782497,
                               cone_radius=0.1775,
                               )
    # test all results are there and the expected target is present
    assert len(result) == MONKEYPATCH_TABLE_LENGTH
    assert 'target_name' in result.colnames
    assert 'b333' in result['target_name']


def test_list_surveys(monkeypatch):
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    saved_list = eso.list_surveys()
    assert isinstance(saved_list, list)
    assert set(TEST_SURVEYS) <= set(saved_list)


def test_list_instruments(monkeypatch):
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    saved_list = eso.list_instruments()
    assert isinstance(saved_list, list)
    assert set(TEST_INSTRUMENTS) <= set(saved_list)


def test_authenticate(monkeypatch):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    monkeypatch.setattr(eso, '_request', eso_request)
    eso.cache_location = DATA_DIR
    authenticated = eso._authenticate(username="someuser", password="somepassword")
    assert authenticated is True


def test_download(monkeypatch, tmp_path):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    eso.cache_location = tmp_path
    fileid = 'testfile'
    filename = os.path.join(tmp_path, f"{fileid}.fits.Z")
    monkeypatch.setattr(eso._session, 'get', download_request)
    downloaded_files = eso.retrieve_data([fileid], unzip=False)
    assert len(downloaded_files) == 1
    assert downloaded_files[0] == filename


@pytest.mark.skipif(sys.platform.startswith("win"), reason="gunzip not available on Windows")
def test_unzip(tmp_path):
    eso = Eso()
    filename = os.path.join(DATA_DIR, 'testfile.fits.Z')
    tmp_filename = tmp_path / 'testfile.fits.Z'
    uncompressed_filename = tmp_path / 'testfile.fits'
    shutil.copy(filename, tmp_filename)
    uncompressed_files = eso._unzip_files([str(tmp_filename)])
    assert len(uncompressed_files) == 1
    assert uncompressed_files[0] == str(uncompressed_filename)


def test_cached_file():
    eso = Eso()
    filename = os.path.join(DATA_DIR, 'testfile.fits.Z')
    assert eso._find_cached_file(filename) is True
    assert eso._find_cached_file("non_existent_filename") is False


def test_calselector(monkeypatch, tmp_path):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    eso.cache_location = tmp_path
    dataset = 'FORS2.2021-01-02T00:59:12.533'
    monkeypatch.setattr(eso._session, 'post', calselector_request)
    result = eso.get_associated_files([dataset], savexml=True)
    assert isinstance(result, list)
    assert len(result) == MONKEYPATCH_TABLE_LENGTH
    assert dataset not in result


def test_calselector_multipart(monkeypatch, tmp_path):
    # monkeypatch instructions from https://pytest.org/latest/monkeypatch.html
    eso = Eso()
    eso.cache_location = tmp_path
    datasets = ['FORS2.2021-01-02T00:59:12.533', 'FORS2.2021-01-02T00:59:12.534']
    monkeypatch.setattr(eso._session, 'post', calselector_request)
    result = eso.get_associated_files(datasets, savexml=False)
    assert isinstance(result, list)
    assert len(result) == 99
    assert datasets[0] not in result and datasets[1] not in result


def test_tap_url():
    prod_url = "https://archive.eso.org/tap_obs"
    eso_instance = Eso()
    assert eso_instance._tap_url() == prod_url


@pytest.mark.parametrize("input_val, expected", [
    # Numeric values
    (1, "= 1"),
    (1.5, "= 1.5"),
    (None, "= None"),

    # String values
    ("ciao", "= 'ciao'"),
    ("1.5", "= '1.5'"),
    ("ciao  ", "= 'ciao'"),
    ("  ciao", "= 'ciao'"),
    ("  ciao  ", "= 'ciao'"),
    ("1.5 ", "= '1.5'"),
    (" a string with spaces ", "= 'a string with spaces'"),
    ("SGR A", "= 'SGR A'"),
    ("'SGR A'", "= 'SGR A'"),

    # Operator-based queries
    ("< 5", "< 5"),
    ("> 1.23", "> 1.23"),
    ("< '5'", "< '5'"),
    ("> '1.23'", "> '1.23'"),

    ("like '%John%'", "like  '%John%'"),
    ("not like '%John%'", "not like  '%John%'"),
    ("in ('apple', 'mango', 'orange')", "in  ('apple', 'mango', 'orange')"),
    ("not in ('apple', 'mango', 'orange')", "not in  ('apple', 'mango', 'orange')"),
    ("in (1, 2, 3)", "in  (1, 2, 3)"),
    ("not in (1, 2, 3)", "not in  (1, 2, 3)"),

    # Operator-based queries
    ("<5", "< 5"),
    (">1.23", "> 1.23"),
    (">=1.23", ">= 1.23"),
    ("<=1.23", "<= 1.23"),
    ("!=1.23", "!= 1.23"),
    ("<'5'", "< '5'"),
    (">'1.23'", "> '1.23'"),

    # Strings that look like operators but should be treated as strings
    ("'like %John%'", "= 'like %John%'"),
    ("'= something'", "= '= something'"),
    ("'> 5'", "= '> 5'"),
    ("' > 1.23 '", "= ' > 1.23 '"),
    ("likewise", "= 'likewise'"),
    ("INfinity", "= 'INfinity'"),
    ("like'%John%'", "= 'like'%John%''"),  # pathologic case

    # Ill-formed queries: Operator, but not sanitized. Expected to be passed through as-is
    ("like %John%", "like  %John%"),
    ("not like %John%", "not like  %John%"),
    ("= SGR A", "= SGR A"),
])
def test_adql_sanitize_op_val(input_val, expected):
    assert _adql_sanitize_op_val(input_val) == expected


def test_maxrec():
    eso_instance = Eso()

    # EXPECTED_MAXREC is the default value in the conf
    maxrec = eso_instance.ROW_LIMIT
    assert maxrec == EXPECTED_MAXREC

    # we change it to 5
    eso_instance.ROW_LIMIT = 5
    maxrec = eso_instance.ROW_LIMIT
    assert maxrec == 5

    # change it to no-truncation
    eso_instance.ROW_LIMIT = None
    maxrec = eso_instance.ROW_LIMIT
    assert maxrec == sys.maxsize

    # no truncation
    eso_instance.ROW_LIMIT = 0
    maxrec = eso_instance.ROW_LIMIT
    assert maxrec == sys.maxsize

    # no truncation
    eso_instance.ROW_LIMIT = -1
    maxrec = eso_instance.ROW_LIMIT
    assert maxrec == sys.maxsize


def test_download_pyvo_table():
    eso_instance = Eso()
    dal = pyvo.dal.TAPService(eso_instance._tap_url())

    q_str = "select * from ivoa.ObsCore"
    table = None
    with pytest.raises(pyvo.dal.exceptions.DALFormatError):
        table = eso_instance._try_download_pyvo_table(q_str, dal)

    assert table is None


def test_issue_table_length_warnings():
    eso_instance = Eso()

    # should warn, since the table is empty
    t = Table()
    with pytest.warns(NoResultsWarning):
        eso_instance._maybe_warn_about_table_length(t, 1)

    # should warn, since EXPECTED_MAXREC = eso_instance.maxrec
    t = Table({"col_name": [i for i in range(EXPECTED_MAXREC+1)]})
    with pytest.warns(MaxResultsWarning):
        eso_instance._maybe_warn_about_table_length(t, EXPECTED_MAXREC+1)

    # should not warn
    t = Table({"col_name": [i for i in range(51)]})
    eso_instance._maybe_warn_about_table_length(t, EXPECTED_MAXREC+1)


def test_reorder_columns(monkeypatch):
    eso = Eso()
    monkeypatch.setattr(eso, 'query_tap', monkey_tap)
    table = eso.query_main(
        column_filters={
            'target': "SGR A",
            'object': "SGR A"}
    )
    names_before = table.colnames[:]
    table2 = _reorder_columns(table)
    names_after = table2.colnames[:]

    # the columns we want to change are actually in the table
    assert set(DEFAULT_LEAD_COLS_RAW).issubset(names_before)
    assert set(DEFAULT_LEAD_COLS_RAW).issubset(names_after)

    # no columns are removed nor added
    assert set(names_before) == set(names_after)

    # Column by column, the contents are the same
    for cname in table.colnames:
        assert table[[cname]].values_equal(table2[[cname]]), f"Error for {cname}"

    # empty table doesn't cause a crash
    empty_1 = Table()
    empty_2 = _reorder_columns(empty_1)
    assert len(empty_1) == 0 and isinstance(empty_1, Table)
    assert len(empty_2) == 0 and isinstance(empty_1, Table)

    # If the values we're looking for as leading columns, everything stays the same
    some_table = Table({"x": [1, 2, 3], "y": [4, 5, 6]})
    same_table = _reorder_columns(some_table)
    assert all(some_table == same_table), "Table with no cols to change fails"

    # If what we pas is not a table, the function has no effect
    not_a_table_1 = object()
    not_a_table_2 = _reorder_columns(not_a_table_1)
    assert not_a_table_1 == not_a_table_2


@pytest.mark.parametrize("params, expected", [
    # Basic SELECT * cases
    (_UserParams(table_name="ivoa.ObsCore"),
     "select * from ivoa.ObsCore"),

    (_UserParams(table_name="ivoa.ObsCore", columns=''),
     "select * from ivoa.ObsCore"),

    (_UserParams(table_name="ivoa.ObsCore", columns='*'),
     "select * from ivoa.ObsCore"),

    # Basic column selection
    (_UserParams(table_name="my.Table", columns="a, b, c"),
     "select a, b, c from my.Table"),

    (_UserParams(table_name="my.Table", columns=["a", "b", "c"]),
     "select a, b, c from my.Table"),

    # With filters and order
    (_UserParams(
        table_name="my.Table",
        columns=["a", "b", "c"],
        column_filters={"x": "> 1", "y": "< 2"},
        order_by="z"),
     "select a, b, c from my.Table where x > 1 and y < 2 order by z desc"),

    # With cone search
    (_UserParams(
        table_name="galaxies",
        columns="name, ra, dec",
        cone_ra=150.1, cone_dec=2.3, cone_radius=0.1),
     "select name, ra, dec from galaxies where "
     "intersects(s_region, circle('ICRS', 150.1, 2.3, 0.1))=1"),

    # With count_only
    (_UserParams(
        table_name="galaxies",
        cone_ra=10, cone_dec=20, cone_radius=0.5,
        count_only=True),
     "select count(*) from galaxies where intersects(s_region, circle('ICRS', 10, 20, 0.5))=1"),

    # With top
    (_UserParams(
        table_name="stars",
        columns="name, mag",
        top=10,
        order_by="mag",
        order_by_desc=False),
     "select top 10 name, mag from stars order by mag asc"),

    # With multiple filters including IN
    (_UserParams(
        table_name="beautiful.Stars",
        columns=["id", "instrument"],
        column_filters={"instrument": "in ('MUSE', 'UVES')", "t_exptime": "> 100"},
        order_by="t_exptime"),
     "select id, instrument from beautiful.Stars where instrument in  ('MUSE', 'UVES') "
     "and t_exptime > 100 order by t_exptime desc"),

    # With all params 1
    (
        _UserParams(
            table_name="ivoa.ObsCore",
            cone_ra=180.0,
            cone_dec=-45.0,
            cone_radius=0.05,
            columns=["target_name", "s_ra", "s_dec", "em_min", "em_max"],
            column_filters={
                "em_min": "> 4e-7",
                "em_max": "< 1.2e-6",
                "dataproduct_type": "in ('spectrum')"
            },
            top=100,
            order_by="em_min",
            order_by_desc=True,
            count_only=False,
            get_query_payload=True,
        ),
        "select top 100 target_name, s_ra, s_dec, em_min, em_max from ivoa.ObsCore "
        "where em_min > 4e-7 and em_max < 1.2e-6 and dataproduct_type in  ('spectrum') "
        "and intersects(s_region, circle('ICRS', 180.0, -45.0, 0.05))=1 "
        "order by em_min desc"
    ),

    # With all params 2
    (
        _UserParams(
            table_name="ivoa.ObsCore",
            cone_ra=180.0,
            cone_dec=-45.0,
            cone_radius=0.05,
            columns=["target_name", "s_ra", "s_dec", "em_min", "em_max"],
            column_filters={
                "em_min": "> 4e-7",
                "em_max": "< 1.2e-6",
                "dataproduct_type": "in ('spectrum')"
            },
            top=100,
            order_by="em_min",
            order_by_desc=False,
            count_only=True,
            get_query_payload=True,
        ),
        "select top 100 count(*) from ivoa.ObsCore "
        "where em_min > 4e-7 and em_max < 1.2e-6 and dataproduct_type in  ('spectrum') "
        "and intersects(s_region, circle('ICRS', 180.0, -45.0, 0.05))=1"
    ),
])
def test_build_adql_string(params, expected):
    """
    Tests that the ADQL query builder generates the expected query.

    Example of typical adql query:

        SELECT
            target_name, dp_id, s_ra, s_dec, t_exptime, em_min, em_max,
            dataproduct_type, instrument_name, obstech, abmaglim,
            proposal_id, obs_collection
        FROM
            ivoa.ObsCore
        WHERE
            intersects(s_region, circle('ICRS', 109.668246, -24.558700, 0.001389)) = 1
            AND dataproduct_type IN ('spectrum')
            AND em_min > 4.0e-7
            AND em_max < 1.2e-6
        ORDER BY
            em_min DESC
    """
    query = _build_adql_string(params)
    assert query == expected, f"Expected:\n{expected}\n\nGot:\n{query}\n"



# Licensed under a 3-clause BSD style license - see LICENSE.rst


import os


def get_package_data():
    paths = [os.path.join('data', '*.csv'),
             os.path.join('data', '*.xml'),
             os.path.join('data', '*.json'),
             os.path.join('data', '*.fits*')
             ]
    return {'salutator.eso.tests': paths}



"""
utils.py: helper functions for the astropy.eso module
"""
from dataclasses import dataclass
from typing import Dict, List, Optional, Union
from astropy.table import Table

DEFAULT_LEAD_COLS_RAW = ['object', 'ra', 'dec', 'dp_id', 'date_obs', 'prog_id']
DEFAULT_LEAD_COLS_PHASE3 = ['target_name', 's_ra', 's_dec', 'dp_id', 'date_obs', 'proposal_id']


@dataclass
class _UserParams:
    """
    Parameters set by the user
    """
    table_name: str
    column_name: str = None
    allowed_values: Union[List[str], str] = None
    cone_ra: float = None
    cone_dec: float = None
    cone_radius: float = None
    columns: Union[List, str] = None
    column_filters: Dict[str, str] = None
    top: int = None
    order_by: str = ''
    order_by_desc: bool = True
    count_only: bool = False
    get_query_payload: bool = False
    print_help: bool = False
    authenticated: bool = False


def _split_str_as_list_of_str(column_str: str):
    if column_str == '':
        column_list = []
    else:
        column_list = list(map(lambda x: x.strip(), column_str.split(',')))
    return column_list


def _raise_if_has_deprecated_keys(filters: Optional[Dict[str, str]]) -> bool:
    if not filters:
        return

    if any(k in filters for k in ("box", "coord1", "coord2")):
        raise ValueError(
            "box, coord1 and coord2 are deprecated; "
            "use cone_ra, cone_dec and cone_radius instead."
        )

    if any(k in filters for k in ("etime", "stime")):
        raise ValueError(
            "'stime' and 'etime' are deprecated; "
            "use instead 'exp_start' together with '<', '>', 'between'. Examples:\n"
            "\tcolumn_filters = {'exp_start': '< 2024-01-01'}\n"
            "\tcolumn_filters = {'exp_start': '>= 2023-01-01'}\n"
            "\tcolumn_filters = {'exp_start': \"between '2023-01-01' and '2024-01-01'\"}\n"
        )


def _build_where_constraints(
        column_name: str,
        allowed_values: Union[List[str], str],
        column_filters: Dict[str, str]) -> str:
    def _format_helper(av):
        if isinstance(av, str):
            av = _split_str_as_list_of_str(av)
        quoted_values = [f"'{v.strip()}'" for v in av]
        return f"{column_name} in ({', '.join(quoted_values)})"

    column_filters = column_filters or {}
    where_constraints = []
    if allowed_values:
        where_constraints.append(_format_helper(allowed_values))

    where_constraints += [
        f"{k} {_adql_sanitize_op_val(v)}" for k, v in column_filters.items()
    ]
    return where_constraints


def _reorder_columns(table: Table,
                     leading_cols: Optional[List[str]] = None):
    """
    Reorders the columns of the pased table so that the
    colums given by the list leading_cols are first.
    If no leading cols are passed, it defaults to
    ['object', 'ra', 'dec', 'dp_id', 'date_obs']
    Returns a table with the columns reordered.
    """
    if not isinstance(table, Table):
        return table

    leading_cols = leading_cols or DEFAULT_LEAD_COLS_RAW
    first_cols = []
    last_cols = table.colnames[:]
    for x in leading_cols:
        if x in last_cols:
            last_cols.remove(x)
            first_cols.append(x)
    last_cols = first_cols + last_cols
    table = table[last_cols]
    return table


def _adql_sanitize_op_val(op_val):
    """
    Expected input:
        "= 5", "< 3.14", "like '%John Doe%'", "in ('item1', 'item2')"
        or just string values like "ESO", "ALMA", "'ALMA'", "John Doe"

    Logic:
        returns "<operator> <value>" if operator is provided.
        Defaults to "= <value>" otherwise.
    """
    supported_operators = ["<=", ">=", "!=", "=", ">", "<",
                           "not like ", "not in ", "not between ",
                           "like ", "between ", "in "]  # order matters

    if not isinstance(op_val, str):
        return f"= {op_val}"

    op_val = op_val.strip()
    for s in supported_operators:
        if op_val.lower().startswith(s):
            operator, value = s, op_val[len(s):].strip()
            return f"{operator} {value}"

    # Default case: no operator. Assign "="
    value = op_val if (op_val.startswith("'") and op_val.endswith("'")) else f"'{op_val}'"
    return f"= {value}"


def raise_if_coords_not_valid(cone_ra: Optional[float] = None,
                              cone_dec: Optional[float] = None,
                              cone_radius: Optional[float] = None) -> bool:
    """
    ra, dec, radius must be either present all three
    or absent all three. Moreover, they must be float
    """
    are_all_none = (cone_ra is None) and (cone_dec is None) and (cone_radius is None)
    are_all_float = isinstance(cone_ra, (float, int)) and \
        isinstance(cone_dec, (float, int)) and \
        isinstance(cone_radius, (float, int))
    is_a_valid_combination = are_all_none or are_all_float
    if not is_a_valid_combination:
        raise ValueError(
            "Either all three (cone_ra, cone_dec, cone_radius) must be present or none.\n"
            "Values provided:\n"
            f"\tcone_ra = {cone_ra}, cone_dec = {cone_dec}, cone_radius = {cone_radius}"
        )


def _build_adql_string(user_params: _UserParams) -> str:
    """
    Return the adql string corresponding to the parameters passed
    See adql examples at https://archive.eso.org/tap_obs/examples
    """
    query_string = None
    columns = user_params.columns or []

    # We assume the coordinates passed are valid
    where_circle = []
    if user_params.cone_radius is not None:
        where_circle += [
            'intersects(s_region, circle(\'ICRS\', '
            f'{user_params.cone_ra}, {user_params.cone_dec}, {user_params.cone_radius}))=1']

    wc = _build_where_constraints(user_params.column_name,
                                  user_params.allowed_values,
                                  user_params.column_filters) + where_circle

    if isinstance(columns, str):
        columns = _split_str_as_list_of_str(columns)
    if columns is None or len(columns) < 1:
        columns = ['*']
    if user_params.count_only:
        columns = ['count(*)']

    # Build the query
    query_string = ', '.join(columns) + ' from ' + user_params.table_name
    if len(wc) > 0:
        where_string = ' where ' + ' and '.join(wc)
        query_string += where_string

    if len(user_params.order_by) > 0 and not user_params.count_only:
        order_string = ' order by ' + user_params.order_by + (' desc ' if user_params.order_by_desc else ' asc ')
        query_string += order_string

    if user_params.top is not None:
        query_string = f"select top {user_params.top} " + query_string
    else:
        query_string = "select " + query_string

    return query_string.strip()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
=====================
ESO Archive Module
=====================
European Southern Observatory (ESO)
"""

import base64
import email
import functools
import json
import os
import os.path
import sys
import re
import shutil
import subprocess
import time
import warnings
import xml.etree.ElementTree as ET
from typing import List, Optional, Tuple, Dict, Set, Union

import astropy.utils.data
import keyring
import requests
from astropy.table import Table, Column
from astropy.utils.decorators import deprecated_renamed_argument
from bs4 import BeautifulSoup
from pyvo.dal import TAPService
from pyvo.dal.exceptions import DALQueryError, DALFormatError

from astroquery import log
from . import conf
from astroquery.exceptions import RemoteServiceError, LoginError, \
    NoResultsWarning, MaxResultsWarning
from astroquery.query import QueryWithLogin
from astroquery.utils import schema
from .utils import _UserParams, raise_if_coords_not_valid, _reorder_columns, \
    _raise_if_has_deprecated_keys, _build_adql_string, \
    DEFAULT_LEAD_COLS_PHASE3, DEFAULT_LEAD_COLS_RAW


class CalSelectorError(Exception):
    """
    Raised on failure to parse CalSelector's response.
    """


class _AuthInfo:
    def __init__(self, username: str, token: str):
        self.username = username
        self.token = token
        self.expiration_time = self._get_exp_time_from_token()

    def _get_exp_time_from_token(self) -> int:
        # "manual" decoding since jwt is not installed
        decoded_token = base64.b64decode(self.token.split(".")[1] + "==")
        return int(json.loads(decoded_token)['exp'])

    def _expired(self) -> bool:
        # we anticipate the expiration time by 10 minutes to avoid issues
        return time.time() > self.expiration_time - 600


class _EsoNames:
    raw_table = "dbo.raw"
    phase3_table = "ivoa.ObsCore"
    raw_instruments_column = "instrument"
    phase3_surveys_column = "obs_collection"

    @staticmethod
    def ist_table(instrument_name):
        """
        Returns the name of the instrument specific table (IST)
        """
        return f"ist.{instrument_name}"

    apex_quicklooks_table = ist_table.__func__("apex_quicklooks")
    apex_quicklooks_pid_column = "project_id"


def unlimited_maxrec(func):
    """
    decorator to overwrite maxrec for specific queries
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not isinstance(self, EsoClass):
            raise ValueError(f"Expecting EsoClass, found {type(self)}")

        tmpvar = self.ROW_LIMIT
        try:
            self.ROW_LIMIT = sys.maxsize
            result = func(self, *args, **kwargs)
        finally:
            self.ROW_LIMIT = tmpvar
        return result
    return wrapper


class EsoClass(QueryWithLogin):
    """
    User facing class to query the ESO archive
    """
    USERNAME = conf.username
    CALSELECTOR_URL = "https://archive.eso.org/calselector/v1/associations"
    DOWNLOAD_URL = "https://dataportal.eso.org/dataPortal/file/"
    AUTH_URL = "https://www.eso.org/sso/oidc/token"
    GUNZIP = "gunzip"

    def __init__(self):
        super().__init__()
        self._auth_info: Optional[_AuthInfo] = None
        self._hash = None
        self._ROW_LIMIT = None
        self.ROW_LIMIT = conf.ROW_LIMIT

    @property
    def ROW_LIMIT(self):
        """
        Getter.
        Safeguard that truncates the number of records returned by a query
        """
        return self._ROW_LIMIT

    @ROW_LIMIT.setter
    def ROW_LIMIT(self, value):
        mr = self._ROW_LIMIT

        # type check
        if not (value is None or isinstance(value, int)):
            raise TypeError(f"ROW_LIMIT attribute must be of type int or None; found {type(value)}")

        if value is None or value < 1:
            mr = sys.maxsize
        else:
            mr = value

        self._ROW_LIMIT = mr

    def _tap_url(self) -> str:
        url = conf.tap_url
        return url

    def _authenticate(self, *, username: str, password: str) -> bool:
        """
        Get the access token from ESO SSO provider
        """
        self._auth_info = None
        url_params = {"response_type": "id_token token",
                      "grant_type": "password",
                      "client_id": "clientid",
                      "client_secret": "clientSecret",
                      "username": username,
                      "password": password}
        log.info(f"Authenticating {username} on 'www.eso.org' ...")
        response = self._request('GET', self.AUTH_URL, params=url_params)
        if response.status_code == 200:
            token = json.loads(response.content)['id_token']
            self._auth_info = _AuthInfo(username=username, token=token)
            log.info("Authentication successful!")
            return True
        else:
            log.error("Authentication failed!")
            return False

    def _get_auth_info(self, username: str, *, store_password: bool = False,
                       reenter_password: bool = False) -> Tuple[str, str]:
        """
        Get the auth info (user, password) for use in another function
        """

        if username is None:
            if not self.USERNAME:
                raise LoginError("If you do not pass a username to login(), "
                                 "you should configure a default one!")
            else:
                username = self.USERNAME

        service_name = "astroquery:www.eso.org"
        # Get password from keyring or prompt
        password, password_from_keyring = self._get_password(
            service_name, username, reenter=reenter_password)

        # When authenticated, save password in keyring if needed
        if password_from_keyring is None and store_password:
            keyring.set_password(service_name, username, password)

        return username, password

    def _login(self, username: str = None, store_password: bool = False,
               reenter_password: bool = False) -> bool:
        """
        Login to the ESO User Portal.

        Parameters
        ----------
        username : str, optional
            Username to the ESO Public Portal. If not given, it should be
            specified in the config file.
        store_password : bool, optional
            Stores the password securely in your keyring. Default is `False`.
        reenter_password : bool, optional
            Asks for the password even if it is already stored in the
            keyring. This is the way to overwrite an already stored passwork
            on the keyring. Default is `False`.
        """

        username, password = self._get_auth_info(username=username,
                                                 store_password=store_password,
                                                 reenter_password=reenter_password)

        return self._authenticate(username=username, password=password)

    def _get_auth_header(self) -> Dict[str, str]:
        if self._auth_info and self._auth_info._expired():
            raise LoginError(
                "Authentication token has expired! Please log in again."
            )
        if self._auth_info and not self._auth_info._expired():
            return {'Authorization': 'Bearer ' + self._auth_info.token}
        else:
            return {}

    def _maybe_warn_about_table_length(self, table_rowlim_plus_one, row_limit_plus_one):
        """
        Issues a warning when a table is empty or when the
        results are truncated
        """
        if len(table_rowlim_plus_one) < 1:
            warnings.warn("Query returned no results", NoResultsWarning)

        if len(table_rowlim_plus_one) == row_limit_plus_one:
            warnings.warn(f"Results truncated to {self.ROW_LIMIT}. "
                          "To retrieve all the records set to None the ROW_LIMIT attribute",
                          MaxResultsWarning)

    def _try_download_pyvo_table(self,
                                 query_str: str,
                                 tap: TAPService) -> Optional[Table]:
        table_with_an_extra_row = Table()

        def message(query_str):
            return (f"Error executing the following query:\n\n"
                    f"{query_str}\n\n"
                    "See examples here: https://archive.eso.org/tap_obs/examples\n\n"
                    f"For maximum query freedom use the query_tap method:\n\n"
                    f' >>> Eso().query_tap( "{query_str}" )\n\n')

        try:
            row_limit_plus_one = self.ROW_LIMIT
            if self.ROW_LIMIT < sys.maxsize:
                row_limit_plus_one = self.ROW_LIMIT + 1

            table_with_an_extra_row = tap.search(query=query_str, maxrec=row_limit_plus_one).to_table()
            self._maybe_warn_about_table_length(table_with_an_extra_row, row_limit_plus_one)
        except DALQueryError:
            log.error(message(query_str))
        except DALFormatError as e:
            raise DALFormatError(message(query_str) + f"cause: {e.cause}") from e
        except Exception as e:
            raise type(e)(f"{e}\n" + message(query_str)) from e

        return table_with_an_extra_row[:self.ROW_LIMIT]

    def tap(self, authenticated: bool = False) -> TAPService:

        if authenticated and not self.authenticated():
            raise LoginError(
                "It seems you are trying to issue an authenticated query without "
                "being authenticated. Possible solutions:\n"
                "1. Set the query function argument authenticated=False"
                " OR\n"
                "2. Login with your username and password: "
                "<eso_class_instance>.login(username=<your_username>"
            )

        log.debug(f"Querying from {self._tap_url()}")
        if authenticated:
            h = self._get_auth_header()
            self._session.headers = {**self._session.headers, **h}
            tap_service = TAPService(self._tap_url(), session=self._session)
        else:
            tap_service = TAPService(self._tap_url())

        return tap_service

    def query_tap(self,
                  query: str, *,
                  authenticated: bool = False,
                  ) -> Table:
        """
        Query the ESO TAP service using a free ADQL string.

        Parameters
        ----------
        query_str : str
            The ADQL query string to be executed.
        authenticated : bool, optional
            If ``True``, the query is run as an authenticated user.
            Authentication must be performed beforehand via
            :meth:`salutator.eso.EsoClass.login`. Authenticated queries
            may be slower. Default is ``False``.

        Returns
        -------
        astropy.table.Table

        Examples
        --------
        from salutator.eso import Eso
        eso_instance = Eso()
        eso_instance.query_tap("SELECT * FROM ivoa.ObsCore")
        """
        table_to_return = Table()
        tap_service = self.tap(authenticated)
        table_to_return = self._try_download_pyvo_table(query, tap_service)
        return table_to_return

    @unlimited_maxrec
    @deprecated_renamed_argument('cache', None, since='0.4.11')
    def list_instruments(self, cache=True) -> List[str]:
        """
        List all the available instrument-specific queries offered by the ESO archive.

        Returns
        -------
        instrument_list : list of strings
        cache : bool
            Deprecated - unused.
        """
        _ = cache  # We're aware about disregarding the argument
        query_str = ("select table_name from TAP_SCHEMA.tables "
                     "where schema_name='ist' order by table_name")
        res = self.query_tap(query_str)["table_name"].data

        l_res = list(res)

        # Remove ist.apex_quicklooks, which is not actually a raw instrument
        if _EsoNames.apex_quicklooks_table in l_res:
            l_res.remove(_EsoNames.apex_quicklooks_table)

        l_res = list(map(lambda x: x.split(".")[1], l_res))

        return l_res

    @unlimited_maxrec
    @deprecated_renamed_argument('cache', None, since='0.4.11')
    def list_surveys(self, *, cache=True) -> List[str]:
        """
        List all the available surveys (phase 3) in the ESO archive.

        Returns
        -------
        collection_list : list of strings
        cache : bool
            Deprecated - unused.
        """
        _ = cache  # We're aware about disregarding the argument
        t = _EsoNames.phase3_table
        c = _EsoNames.phase3_surveys_column
        query_str = f"select distinct {c} from {t}"
        res = list(self.query_tap(query_str)[c].data)
        return res

    @unlimited_maxrec
    def list_column(self, table_name: str) -> None:
        """
        Prints the columns contained in a given table
        """
        help_query = (
            f"select column_name, datatype, xtype, unit "
            # TODO: The column description renders output unmanageable
            # f", description "
            f"from TAP_SCHEMA.columns "
            f"where table_name = '{table_name}'")
        available_cols = self.query_tap(help_query)

        count_query = f"select count(*) from {table_name}"
        num_records = list(self.query_tap(count_query)[0].values())[0]

        with (astropy.conf.set_temp(
                "max_lines", len(available_cols) + 2),
                astropy.conf.set_temp(
                "max_width", sys.maxsize)):
            log.info(f"\nColumns present in the table {table_name}:\n{available_cols}\n"
                     f"\nNumber of records present in the table {table_name}:\n{num_records}\n")

    def _query_on_allowed_values(
        self,
        user_params: _UserParams
    ) -> Union[Table, int, str, None]:
        if user_params.print_help:
            self.list_column(user_params.table_name)
            return

        _raise_if_has_deprecated_keys(user_params.column_filters)

        raise_if_coords_not_valid(user_params.cone_ra, user_params.cone_dec, user_params.cone_radius)

        query = _build_adql_string(user_params)

        if user_params.get_query_payload:
            return query

        ret_table = self.query_tap(query=query, authenticated=user_params.authenticated)
        return list(ret_table[0].values())[0] if user_params.count_only else ret_table

    @deprecated_renamed_argument(('open_form', 'cache'), (None, None),
                                 since=['0.4.11', '0.4.11'])
    def query_surveys(
            self,
            surveys: Union[List[str], str] = None, *,
            cone_ra: float = None, cone_dec: float = None, cone_radius: float = None,
            columns: Union[List, str] = None,
            column_filters: Optional[dict] = None,
            top: int = None,
            count_only: bool = False,
            get_query_payload: bool = False,
            help: bool = False,
            authenticated: bool = False,
            open_form: bool = False, cache: bool = False,
    ) -> Union[Table, int, str]:
        """
        Query survey Phase 3 data contained in the ESO archive.

        Parameters
        ----------
        surveys : str or list, optional
            Name of the survey(s) to query. Should be one or more of the
            names returned by :meth:`~salutator.eso.EsoClass.list_surveys`. If specified
            as a string, it should be a comma-separated list of survey names.
            If not specified, returns records relative to all surveys. Default is ``None``.
        cone_ra : float, optional
            Cone Search Center - Right Ascension in degrees.
        cone_dec : float, optional
            Cone Search Center - Declination in degrees.
        cone_radius : float, optional
            Cone Search Radius in degrees.
        columns : str or list of str, optional
            Name of the columns the query should return. If specified as a string,
            it should be a comma-separated list of column names.
        top : int, optional
            When set to ``N``, returns only the top ``N`` records.
        count_only : bool, optional
            If ``True``, returns only an ``int``: the count of the records
            the query would return when set to ``False``. Default is ``False``.
        get_query_payload : bool, optional
            If ``True``, returns only a ``str``: the query string that
            would be issued to the TAP service. Default is ``False``.
        help : bool, optional
            If ``True``, prints all the parameters accepted in ``column_filters``
            and ``columns``. Default is ``False``.
        authenticated : bool, optional
            If ``True``, runs the query as an authenticated user.
            Authentication must be done beforehand via
            :meth:`~salutator.eso.EsoClass.login`. Authenticated queries may be slower.
            Default is ``False``.
        column_filters : dict or None, optional
            Constraints applied to the query in ADQL syntax,
            e.g., ``{"exp_start": "between '2024-12-31' and '2025-12-31'"}``.
            Default is ``None``.
        open_form : bool, optional
            **Deprecated** - unused.
        cache : bool, optional
            **Deprecated** - unused.

        Returns
        -------
        astropy.table.Table, str, int or None
            - By default, returns an :class:`~astropy.table.Table` containing records
              based on the specified columns and constraints.
              Returns ``None`` if the query has no results.
            - When ``count_only`` is ``True``, returns an ``int`` representing the
              record count for the specified filters.
            - When ``get_query_payload`` is ``True``, returns the query string that
              would be issued to the TAP service given the specified arguments.
        """
        _ = open_form, cache  # make explicit that we are aware these arguments are unused
        column_filters = column_filters if column_filters else {}
        user_params = _UserParams(table_name=_EsoNames.phase3_table,
                                  column_name=_EsoNames.phase3_surveys_column,
                                  allowed_values=surveys,
                                  cone_ra=cone_ra,
                                  cone_dec=cone_dec,
                                  cone_radius=cone_radius,
                                  columns=columns,
                                  column_filters=column_filters,
                                  top=top,
                                  count_only=count_only,
                                  get_query_payload=get_query_payload,
                                  print_help=help,
                                  authenticated=authenticated,
                                  )
        t = self._query_on_allowed_values(user_params=user_params)
        t = _reorder_columns(t, DEFAULT_LEAD_COLS_PHASE3)
        return t

    @deprecated_renamed_argument(('open_form', 'cache'), (None, None),
                                 since=['0.4.11', '0.4.11'])
    def query_main(
            self,
            instruments: Union[List[str], str] = None, *,
            cone_ra: float = None, cone_dec: float = None, cone_radius: float = None,
            columns: Union[List, str] = None,
            column_filters: Optional[dict] = None,
            top: int = None,
            count_only: bool = False,
            get_query_payload: bool = False,
            help: bool = False,
            authenticated: bool = False,
            open_form: bool = False, cache: bool = False,
    ) -> Union[Table, int, str]:
        """
        Query raw data from all instruments contained in the ESO archive.

        Parameters
        ----------
        instruments : str or list, optional
            Name of the instruments to filter. Should be one or more of the
            names returned by :meth:`~salutator.eso.EsoClass.list_instruments`. If specified
            as a string, it should be a comma-separated list of instrument names.
            If not specified, returns records relative to all instruments. Default is ``None``.
        cone_ra : float, optional
            Cone Search Center - Right Ascension in degrees.
        cone_dec : float, optional
            Cone Search Center - Declination in degrees.
        cone_radius : float, optional
            Cone Search Radius in degrees.
        columns : str or list of str, optional
            Name of the columns the query should return. If specified as a string,
            it should be a comma-separated list of column names.
        top : int, optional
            When set to ``N``, returns only the top ``N`` records.
        count_only : bool, optional
            If ``True``, returns only an ``int``: the count of the records
            the query would return when set to ``False``. Default is ``False``.
        get_query_payload : bool, optional
            If ``True``, returns only a ``str``: the query string that
            would be issued to the TAP service. Default is ``False``.
        help : bool, optional
            If ``True``, prints all the parameters accepted in ``column_filters``
            and ``columns``. Default is ``False``.
        authenticated : bool, optional
            If ``True``, runs the query as an authenticated user.
            Authentication must be done beforehand via
            :meth:`~salutator.eso.EsoClass.login`. Authenticated queries may be slower.
            Default is ``False``.
        column_filters : dict or None, optional
            Constraints applied to the query in ADQL syntax,
            e.g., ``{"exp_start": "between '2024-12-31' and '2025-12-31'"}``.
            Default is ``None``.
        open_form : bool, optional
            **Deprecated** - unused.
        cache : bool, optional
            **Deprecated** - unused.

        Returns
        -------
        astropy.table.Table, str, int or None
            - By default, returns an :class:`~astropy.table.Table` containing records
              based on the specified columns and constraints.
              Returns ``None`` if the query has no results.
            - When ``count_only`` is ``True``, returns an ``int`` representing the
              record count for the specified filters.
            - When ``get_query_payload`` is ``True``, returns the query string that
              would be issued to the TAP service given the specified arguments.
        """
        _ = open_form, cache  # make explicit that we are aware these arguments are unused
        column_filters = column_filters if column_filters else {}
        user_params = _UserParams(table_name=_EsoNames.raw_table,
                                  column_name=_EsoNames.raw_instruments_column,
                                  allowed_values=instruments,
                                  cone_ra=cone_ra,
                                  cone_dec=cone_dec,
                                  cone_radius=cone_radius,
                                  columns=columns,
                                  column_filters=column_filters,
                                  top=top,
                                  count_only=count_only,
                                  get_query_payload=get_query_payload,
                                  print_help=help,
                                  authenticated=authenticated,
                                  )
        t = self._query_on_allowed_values(user_params)
        t = _reorder_columns(t, DEFAULT_LEAD_COLS_RAW)
        return t

    @deprecated_renamed_argument(('open_form', 'cache'), (None, None),
                                 since=['0.4.11', '0.4.11'])
    def query_instrument(
            self,
            instrument: str, *,
            cone_ra: float = None, cone_dec: float = None, cone_radius: float = None,
            columns: Union[List, str] = None,
            column_filters: Optional[dict] = None,
            top: int = None,
            count_only: bool = False,
            get_query_payload: bool = False,
            help: bool = False,
            authenticated: bool = False,
            open_form: bool = False, cache: bool = False,
    ) -> Union[Table, int, str]:
        """
        Query instrument-specific raw data contained in the ESO archive.

        Parameters
        ----------
        instrument : str
            Name of the instrument from which raw data is to be queried.
            Should be ONLY ONE of the names returned by
            :meth:`~salutator.eso.EsoClass.list_instruments`.
        cone_ra : float, optional
            Cone Search Center - Right Ascension in degrees.
        cone_dec : float, optional
            Cone Search Center - Declination in degrees.
        cone_radius : float, optional
            Cone Search Radius in degrees.
        columns : str or list of str, optional
            Name of the columns the query should return. If specified as a string,
            it should be a comma-separated list of column names.
        top : int, optional
            When set to ``N``, returns only the top ``N`` records.
        count_only : bool, optional
            If ``True``, returns only an ``int``: the count of the records
            the query would return when set to ``False``. Default is ``False``.
        get_query_payload : bool, optional
            If ``True``, returns only a ``str``: the query string that
            would be issued to the TAP service. Default is ``False``.
        help : bool, optional
            If ``True``, prints all the parameters accepted in ``column_filters``
            and ``columns``. Default is ``False``.
        authenticated : bool, optional
            If ``True``, runs the query as an authenticated user.
            Authentication must be done beforehand via
            :meth:`~salutator.eso.EsoClass.login`. Note that authenticated queries are slower.
            Default is ``False``.
        column_filters : dict or None, optional
            Constraints applied to the query in ADQL syntax,
            e.g., ``{"exp_start": "between '2024-12-31' and '2025-12-31'"}``.
            Default is ``None``.
        open_form : bool, optional
            **Deprecated** - unused.
        cache : bool, optional
            **Deprecated** - unused.

        Returns
        -------
        astropy.table.Table, str, int, or None
            - By default, returns an :class:`~astropy.table.Table` containing records
              based on the specified columns and constraints. Returns ``None`` if no results.
            - When ``count_only`` is ``True``, returns an ``int`` representing the
              record count for the specified filters.
            - When ``get_query_payload`` is ``True``, returns the query string that
              would be issued to the TAP service given the specified arguments.
        """
        _ = open_form, cache  # make explicit that we are aware these arguments are unused
        column_filters = column_filters if column_filters else {}
        user_params = _UserParams(table_name=_EsoNames.ist_table(instrument),
                                  column_name=None,
                                  allowed_values=None,
                                  cone_ra=cone_ra,
                                  cone_dec=cone_dec,
                                  cone_radius=cone_radius,
                                  columns=columns,
                                  column_filters=column_filters,
                                  top=top,
                                  count_only=count_only,
                                  get_query_payload=get_query_payload,
                                  print_help=help,
                                  authenticated=authenticated)
        t = self._query_on_allowed_values(user_params)
        t = _reorder_columns(t, DEFAULT_LEAD_COLS_RAW)
        return t

    def get_headers(self, product_ids, *, cache=True):
        """
        Get the headers associated to a list of data product IDs

        This method returns a `~astropy.table.Table` where the rows correspond
        to the provided data product IDs, and the columns are from each of
        the Fits headers keywords.

        Note: The additional column ``'DP.ID'`` found in the returned table
        corresponds to the provided data product IDs.

        Parameters
        ----------
        product_ids : either a list of strings or a `~astropy.table.Column`
            List of data product IDs.
        cache : bool
            Defaults to True. If set overrides global caching behavior.
            See :ref:`caching documentation <astroquery_cache>`.

        Returns
        -------
        result : `~astropy.table.Table`
            A table where: columns are header keywords, rows are product_ids.

        """
        _schema_product_ids = schema.Schema(
            schema.Or(Column, [schema.Schema(str)]))
        _schema_product_ids.validate(product_ids)
        # Get all headers
        result = []
        for dp_id in product_ids:
            response = self._request(
                "GET", f"https://archive.eso.org/hdr?DpId={dp_id}",
                cache=cache)
            root = BeautifulSoup(response.content, 'html5lib')
            hdr = root.select('pre')[0].text
            header = {'DP.ID': dp_id}
            for key_value in hdr.split('\n'):
                if "=" in key_value:
                    key, value = key_value.split('=', 1)
                    key = key.strip()
                    value = value.split('/', 1)[0].strip()
                    if key[0:7] != "COMMENT":  # drop comments
                        if value == "T":  # Convert boolean T to True
                            value = True
                        elif value == "F":  # Convert boolean F to False
                            value = False
                        # Convert to string, removing quotation marks
                        elif value[0] == "'":
                            value = value[1:-1]
                        elif "." in value:  # Convert to float
                            value = float(value)
                        else:  # Convert to integer
                            value = int(value)
                        header[key] = value
                elif key_value.startswith("END"):
                    break
            result.append(header)
        # Identify all columns
        columns = []
        column_types = []
        for header in result:
            for key in header:
                if key not in columns:
                    columns.append(key)
                    column_types.append(type(header[key]))
        # Add all missing elements
        for item in result:
            for (column, column_type) in zip(columns, column_types):
                if column not in item:
                    item[column] = column_type()
        # Return as Table
        return Table(result)

    @staticmethod
    def _get_filename_from_response(response: requests.Response) -> str:
        content_disposition = response.headers.get("Content-Disposition", "")
        filename = re.findall(r"filename=(\S+)", content_disposition)
        if not filename:
            raise RemoteServiceError(f"Unable to find filename for {response.url}")
        return os.path.basename(filename[0].replace('"', ''))

    @staticmethod
    def _find_cached_file(filename: str) -> bool:
        files_to_check = [filename]
        if filename.endswith(('fits.Z', 'fits.gz')):
            files_to_check.append(filename.rsplit(".", 1)[0])
        for file in files_to_check:
            if os.path.exists(file):
                log.info(f"Found cached file {file}")
                return True
        return False

    def _download_eso_file(self, file_link: str, destination: str,
                           overwrite: bool) -> Tuple[str, bool]:
        block_size = astropy.utils.data.conf.download_block_size
        headers = self._get_auth_header()
        with self._session.get(file_link, stream=True, headers=headers) as response:
            response.raise_for_status()
            filename = self._get_filename_from_response(response)
            filename = os.path.join(destination, filename)
            part_filename = filename + ".part"
            if os.path.exists(part_filename):
                log.info(f"Removing partially downloaded file {part_filename}")
                os.remove(part_filename)
            download_required = overwrite or not self._find_cached_file(filename)
            if download_required:
                with open(part_filename, 'wb') as fd:
                    for chunk in response.iter_content(chunk_size=block_size):
                        fd.write(chunk)
                os.rename(part_filename, filename)
        return filename, download_required

    def _download_eso_files(self, file_ids: List[str], destination: Optional[str],
                            overwrite: bool) -> List[str]:
        destination = destination or self.cache_location
        destination = os.path.abspath(destination)
        os.makedirs(destination, exist_ok=True)
        nfiles = len(file_ids)
        log.info(f"Downloading {nfiles} files ...")
        downloaded_files = []
        for i, file_id in enumerate(file_ids, 1):
            file_link = self.DOWNLOAD_URL + file_id
            log.info(f"Downloading file {i}/{nfiles} {file_link} to {destination}")
            try:
                filename, downloaded = self._download_eso_file(file_link, destination, overwrite)
                downloaded_files.append(filename)
                if downloaded:
                    log.info(f"Successfully downloaded dataset {file_id} to {filename}")
            except requests.HTTPError as http_error:
                if http_error.response.status_code == 401:
                    log.error(f"Access denied to {file_link}")
                else:
                    log.error(f"Failed to download {file_link}. {http_error}")
            except RuntimeError as ex:
                log.error(f"Failed to download {file_link}. {ex}")
        return downloaded_files

    def _unzip_file(self, filename: str) -> str:
        """
        Uncompress the provided file with gunzip.

        Note: ``system_tools.gunzip`` does not work with .Z files
        """
        uncompressed_filename = None
        if filename.endswith(('fits.Z', 'fits.gz')):
            uncompressed_filename = filename.rsplit(".", 1)[0]
            if not os.path.exists(uncompressed_filename):
                log.info(f"Uncompressing file {filename}")
                try:
                    subprocess.run([self.GUNZIP, filename], check=True)
                except Exception as ex:
                    uncompressed_filename = None
                    log.error(f"Failed to unzip {filename}: {ex}")
        return uncompressed_filename or filename

    def _unzip_files(self, files: List[str]) -> List[str]:
        if shutil.which(self.GUNZIP):
            files = [self._unzip_file(file) for file in files]
        else:
            warnings.warn("Unable to unzip files "
                          "(gunzip is not available on this system)")
        return files

    @staticmethod
    def _get_unique_files_from_association_tree(xml: str) -> Set[str]:
        tree = ET.fromstring(xml)
        return {element.attrib['name'] for element in tree.iter('file')}

    def _save_xml(self, payload: bytes, filename: str, destination: str):
        destination = destination or self.cache_location
        destination = os.path.abspath(destination)
        os.makedirs(destination, exist_ok=True)
        filename = os.path.join(destination, filename)
        log.info(f"Saving Calselector association tree to {filename}")
        with open(filename, "wb") as fd:
            fd.write(payload)

    def get_associated_files(self, datasets: List[str], *, mode: str = "raw",
                             savexml: bool = False, destination: str = None) -> List[str]:
        """
        Invoke Calselector service to find calibration files associated to the provided datasets.

        Parameters
        ----------
        datasets : list of strings
            List of datasets for which calibration files should be retrieved.
        mode : string
            Calselector mode: 'raw' (default) for raw calibrations,
             or 'processed' for processed calibrations.
        savexml : bool
            If true, save to disk the XML association tree returned by Calselector.
        destination : string
            Directory where the XML files are saved (default = astropy cache).

        Returns
        -------
        files :
            List of unique datasets associated to the input datasets.
        """
        mode = "Raw2Raw" if mode == "raw" else "Raw2Master"
        post_data = {"dp_id": datasets, "mode": mode}
        response = self._session.post(self.CALSELECTOR_URL, data=post_data)
        response.raise_for_status()
        associated_files = set()
        content_type = response.headers['Content-Type']
        # Calselector can return one or more XML association trees depending on the input.
        # For a single dataset it returns one XML and content-type = 'application/xml'
        if 'application/xml' in content_type:
            filename = self._get_filename_from_response(response)
            xml = response.content
            associated_files.update(self._get_unique_files_from_association_tree(xml))
            if savexml:
                self._save_xml(xml, filename, destination)
        # For multiple datasets it returns a multipart message
        elif 'multipart/form-data' in content_type:
            msg = email.message_from_string(
                f'Content-Type: {content_type}\r\n' + response.content.decode())
            for part in msg.get_payload():
                filename = part.get_filename()
                xml = part.get_payload(decode=True)
                associated_files.update(self._get_unique_files_from_association_tree(xml))
                if savexml:
                    self._save_xml(xml, filename, destination)
        else:
            raise CalSelectorError(f"Unexpected content-type '{content_type}' for {response.url}")

        # remove input datasets from calselector results
        return list(associated_files.difference(set(datasets)))

    def retrieve_data(self, datasets, *, continuation=False, destination=None,
                      with_calib=None, unzip=True):
        """
        Retrieve a list of datasets form the ESO archive.

        Parameters
        ----------
        datasets : list of strings or string
            List of datasets strings to retrieve from the archive.
        destination: string
            Directory where the files are copied.
            Files already found in the destination directory are skipped,
            unless continuation=True.
            Default to astropy cache.
        continuation : bool
            Force the retrieval of data that are present in the destination
            directory.
        with_calib : string
            Retrieve associated calibration files: None (default), 'raw' for
            raw calibrations, or 'processed' for processed calibrations.
        unzip : bool
            Unzip compressed files from the archive after download. `True` by
            default.

        Returns
        -------
        files : list of strings or string
            List of files that have been locally downloaded from the archive.

        Examples
        --------
        dptbl = Eso.query_instrument('apex', pi_coi='ginsburg')
        dpids = [row['DP.ID'] for row in dptbl if 'Map' in row['Object']]
        files = Eso.retrieve_data(dpids)

        """
        return_string = False
        if isinstance(datasets, str):
            return_string = True
            datasets = [datasets]
        if isinstance(datasets, Column):
            datasets = list(datasets)

        if with_calib and with_calib not in ('raw', 'processed'):
            raise ValueError("Invalid value for 'with_calib'. "
                             "It must be 'raw' or 'processed'")

        associated_files = []
        if with_calib:
            log.info(f"Retrieving associated '{with_calib}' calibration files ...")
            try:
                # batch calselector requests to avoid possible issues on the ESO server
                batch_size = 100
                sorted_datasets = sorted(datasets)
                for i in range(0, len(sorted_datasets), batch_size):
                    associated_files += self.get_associated_files(
                        sorted_datasets[i:i + batch_size], mode=with_calib)
                associated_files = list(set(associated_files))
                log.info(f"Found {len(associated_files)} associated files")
            except Exception as ex:
                log.error(f"Failed to retrieve associated files: {ex}")

        all_datasets = datasets + associated_files
        log.info("Downloading datasets ...")
        files = self._download_eso_files(all_datasets, destination, continuation)
        if unzip:
            files = self._unzip_files(files)
        log.info("Done!")
        return files[0] if files and len(files) == 1 and return_string else files

    @deprecated_renamed_argument(('open_form', 'cache'), (None, None),
                                 since=['0.4.11', '0.4.11'])
    def query_apex_quicklooks(self,
                              project_id: Union[List[str], str] = None, *,
                              columns: Union[List, str] = None,
                              column_filters: Optional[dict] = None,
                              top: int = None,
                              count_only: bool = False,
                              get_query_payload: bool = False,
                              help: bool = False,
                              authenticated: bool = False,
                              open_form: bool = False, cache: bool = False,
                              ) -> Union[Table, int, str]:
        """
        APEX data are distributed with quicklook products identified with a
        different name than other ESO products. This query tool searches by
        project ID or any other supported keywords.

        Parameters
        ----------
        project_id : str
            ID of the project from which APEX quicklook data is to be queried.
        columns : str or list of str, optional
            Name of the columns the query should return. If specified as a string,
            it should be a comma-separated list of column names.
        top : int, optional
            When set to ``N``, returns only the top ``N`` records.
        count_only : bool, optional
            If ``True``, returns only an ``int``: the count of the records
            the query would return when set to ``False``. Default is ``False``.
        get_query_payload : bool, optional
            If ``True``, returns only a ``str``: the query string that
            would be issued to the TAP service. Default is ``False``.
        help : bool, optional
            If ``True``, prints all the parameters accepted in ``column_filters``
            and ``columns``. Default is ``False``.
        authenticated : bool, optional
            If ``True``, runs the query as an authenticated user.
            Authentication must be done beforehand via
            :meth:`~salutator.eso.EsoClass.login`. Note that authenticated queries
            are slower. Default is ``False``.
        column_filters : dict or None, optional
            Constraints applied to the query in ADQL syntax,
            e.g., ``{"exp_start": "between '2024-12-31' and '2025-12-31'"}``.
            Default is ``None``.
        open_form : bool, optional
            **Deprecated** - unused.
        cache : bool, optional
            **Deprecated** - unused.

        Returns
        -------
        astropy.table.Table, str, int, or None
            - By default, returns an :class:`~astropy.table.Table` containing records
              based on the specified columns and constraints. Returns ``None`` if no results.
            - When ``count_only`` is ``True``, returns an ``int`` representing the
              record count for the specified filters.
            - When ``get_query_payload`` is ``True``, returns the query string that
              would be issued to the TAP service given the specified arguments.
        """
        _ = open_form, cache  # make explicit that we are aware these arguments are unused
        column_filters = column_filters if column_filters else {}
        user_params = _UserParams(table_name=_EsoNames.apex_quicklooks_table,
                                  column_name=_EsoNames.apex_quicklooks_pid_column,
                                  allowed_values=project_id,
                                  cone_ra=None, cone_dec=None, cone_radius=None,
                                  columns=columns,
                                  column_filters=column_filters,
                                  top=top,
                                  count_only=count_only,
                                  get_query_payload=get_query_payload,
                                  print_help=help,
                                  authenticated=authenticated)
        return self._query_on_allowed_values(user_params)


Eso = EsoClass()



# tests/test_salutator.py

from salutator.greetings import Greeter
from salutator.goodbyes import GoodByer

def test_greeter():
    g = Greeter(name='John')
    g.salute()  # prints greeting

def test_goodbyer():
    gb = GoodByer(name='Diane')
    gb.salute()  # prints goodbye

